/*-
 * A more complex program: each device is a pipe-like device where
 * read() and write() get and put messages.  The delimiters are
 * preserved, so the device is appropriate for buffering protocols.
 * 
 * A read() on an fd not opened O_NONBLOCK will wait until a message
 * arrives at the device before returning.
 *
 */

#include <stdlib.h>
#include <malloc.h>
#include <libc.h>
#include "iomgr.h"
#include "daq.h"

#include <sys/io_msg.h>
#include <sys/kernel.h>
#include <sys/uio.h>
#include <sys/inline.h>
#include <sys/irqinfo.h>
#include <sys/kernel.h>
#include <sys/proxy.h>
#include "daq.h"

extern int DQVerbose;



/* global variables to share with the interrupt handler */
struct dqIrProxy DQIrProxyList[DAQ_IR_MAX_PROXY];
int DQIrProxyCount=0;

int DQCreateDev(inode_t **ip)
{
	inode_t     	*io;
	dqDev		*dev;
	static  iofuns 	pclass;

	if (!pclass.write) {
		pclass.read  = DQRead;
		pclass.close  = DQClose;
		pclass.qioctl  = DQQioctl;
	}

	if (DQIrProxyCount >= DAQ_IR_MAX_PROXY) {
		if (DQVerbose) printf("Cannot support Any more devices \n");
		return ENOMEM;
	}
	

	if ((dev = calloc(sizeof(dqDev),1)) == 0) {
		if (DQVerbose) printf("Calloc of dev failed\n");
		return 0;
	}

	DQIrProxyList[DQIrProxyCount].dev = dev;


	dev->irFreq = DAQ_IR_FREQ_DEFAULT;
	dev->irFreqInterimCnt = 0;
	dev->lowFilter = DAQ_IR_LOW_FILTER;
	dev->highFilter = DAQ_IR_HIGH_FILTER;
	dev->curChar = DAQ_IR_CUR_CHAR;
	dev->overFlow = 0;
	dev->oldIrReadBuf = 2;


	dev->buf[0].len = DAQ_IR_READ_BUF_SIZE;
	dev->buf[1].len = DAQ_IR_READ_BUF_SIZE;
	


	/* Install proxy for ir handler for this device */
	DQIrProxyList[DQIrProxyCount].proxy = qnx_proxy_attach(0, NULL, 0, -1);
	if (DQVerbose) printf("Proxy Id is %d\n", 
		DQIrProxyList[DQIrProxyCount].proxy);
	

	/* Only setup the interrupt handler once. The first time. */
	if (DQIrProxyCount == 0)
	{
		if (DQVerbose) printf("Installing hardware interrupt\n");

		/* Install hardware Interrupt. Share data from my_ds() */
		DQIrProxyList[DQIrProxyCount].hintid =
			qnx_hint_attach(0, DQHint, my_ds());

	}

	/* Setup the proxy  callback. */
	if (server_addpid(DQIrProxyList[DQIrProxyCount].proxy, DQHintProxy) != 0)
	{
		if (DQVerbose) printf("Failed adding server pid\n");
		free(dev);
		qnx_proxy_detach(DQIrProxyList[DQIrProxyCount].proxy);
		if (DQIrProxyCount == 0)
		{
			qnx_hint_detach(DQIrProxyList[DQIrProxyCount].hintid);
		}
		return ENOMEM;
	}

	if ((io=inode_alloc(&pclass)) == 0) 
	{
		if (DQVerbose) printf("Failed performing the inode_alloc\n");
		free(dev);
		qnx_proxy_detach(DQIrProxyList[DQIrProxyCount].proxy);
		if (DQIrProxyCount == 0)
		{
			qnx_hint_detach(DQIrProxyList[DQIrProxyCount].hintid);
		}
		return ENOMEM;
	}

	
	DQIrProxyCount++;
	pclass.minor++;
	io->ino_ftype |= S_IFCHR;
	io->ino_data = dev;
	*ip = io;
	return EOK;
}
