/*-
 * A basic 'main' program for the io manager.
 *
 */


#include <libc.h>
#include <sys/stat.h>
#include "iointernal.h"

int verbose;
char    *progname = "null";
static int ndevs = 8;
char      *dpath = "/testing";

char msg[256];

fatalerr(const char *fmt, ...)
{
	va_list	v;
	va_start(v, fmt);
	write(STDERR_FILENO, progname, strlen(progname));
	write(STDERR_FILENO, ": ", 2);
	vsprintf(msg, fmt, v);
	write(STDERR_FILENO, msg, strlen(msg));
	va_end(v);
	exit(EXIT_FAILURE);
}

void
panic(const char *msg)
{
	sprintf(msg,"%s: panic '%s'\n", progname, msg);
	write(STDERR_FILENO, msg, strlen(msg));
	exit(EXIT_FAILURE);
}
/*
 * FUNCTION: setup
 *
 * DESCRIPTION:
 *	This function will:
 *		- start the initialization of the I/O Manager library 
 *		- register the pathname 
 *		- invoke the I/O Manager library server to process all
 *		  incoming messages. Should never return from server process.
 *		- will never return.
 * 	The root (inode_t) variable will contain a pointer to an inode for this
 *	I/O manager. The ino_data field of this structure will contain an 
 *	array of directories which consist of a pathname and inode pair.
 *
 *
 * INPUT: none
 * 
 * OUTPUT: none
 * 
 * RETURN: never
 */

setup()
{
	int       err;
	inode_t   *root;
	int        create_dev(inode_t **ip);
	dev_t      rm;

	if (server_init() == -1) {
		fatalerr("Unable to become server '%s'\n", strerror(errno));
	}
	if ((err=inode_init(ndevs)) != EOK) {
		fatalerr("unable to init inode table (%s)\n", strerror(err));
	}

	root = create_dir(create_dev);
	d_link(root, ".", root);
	d_link(root, "..", root);
	if ((errno=root_mount(dpath, root)) != EOK) {
		fatalerr("cannot create root dir '%s'\n", strerror(errno));
	}

	if ((rm=getunit(0)) == -1) {
		fatalerr("cannot get root entry\n");
	}
	server();
	exit(0);
}


/*
 * FUNCTION: main
 * 
 * DESCRIPTION:
 *	This is the main entry point into the I/O Manager framework.
 *	This function will:
 *		- parse the input paramters.
 *		- call setup to setup the I/O Manager
 *		- fork if the program is started with a specified pathname
 *		  of subpaths. If the fork is called then the child will
 *		  continue on with the setup function and the parent will
 *		  wait for the new pathname to be registered then it will
 *		  use the open function to create all subpaths provided on
 *		  the command line and then exit. 
 *		  
 *	
 *
 * USEAGE:
 *		prog [-v][-nN][-ppath] [subpatha][subpathb][...]
 *		where:
 *			-v = verbose output. Default No verbosity
 *			-n = N number of devices to allocate for. Default 8.
 *			-p = pathname starting with / used to register this
 *			     process to. The subpaths are optional sub-names
 *			     under path. (ie. initial devices )
 *			     If pathname is not specified then a default
 *			     pathname /testing will be created. 
 *
 *		example: prog -v -n32 -p /daq daq1 daq2
 *			will set verbose on, device structure to max 32,
 *			create pathname /daq and attach this process to it,
 *			create subnames /daq/daq1 and /daq/daq2.
 *
 * 
 *
 */
main(int argc, char **argv)
{
	int       c;

	progname = argv[0];
	while ((c=getopt(argc, argv, "vn:p:")) != -1) {
		switch (c) {
			/* Check and set verbose global var */
		case 'v': verbose = 1; break;
			/* Overide the default number of devices */
		case 'n': ndevs = strtol(optarg,0,0); break;
			/* Get the desired path and any sub-paths */
		case 'p': dpath = optarg; break;
		}
	}

	if (optind < argc) {
		if (fork() == 0) {
			signal(SIGHUP, SIG_IGN);
			signal(SIGINT, SIG_IGN);
			setup();
			exit(0);
		} else {
			struct stat st;
			while (stat(dpath, &st) == -1) {
				sleep(1);
			}
			for (c=optind; c < argc; c++) {
				char  path[256];
				int   fd;
				strcpy(path, dpath);
				strcat(path, "/");
				strcat(path, argv[c]);
				if ((fd=open(path, O_CREAT|O_EXCL,0775))==-1) {
					fatalerr("cannot create %s (%s)\n",
						path, strerror(errno));
				}
				close(fd);
			}
			exit(0);
		}
	} else {
		setup();
		exit(0);
	}
}
