/*-
 * A more complex program: each device is a pipe-like device where
 * read() and write() get and put messages.  The delimiters are
 * preserved, so the device is appropriate for buffering protocols.
 * 
 * A read() on an fd not opened O_NONBLOCK will wait until a message
 * arrives at the device before returning.
 *
 */

#include <libc.h>
#include "iomgr.h"

#include <sys/io_msg.h>
#include <sys/kernel.h>
#include <sys/uio.h>

typedef struct Msg   Msg;
typedef struct Queue Queue;
typedef struct Proc  Proc;

static Queue *Qlist;


struct Queue {
	Msg          *head;
	Msg          *tail;
	Proc         *waiting;
	Queue        *qnext;
};

struct Msg {
	Msg          *next;
	unsigned      nb;
	char         *data;
};

struct Proc {
	pid_t         pid;
	Proc         *next;
	long          size;
};

static Msg *
allocb(unsigned n)
{
	Msg   *b;
	if (b=malloc(sizeof *b)) {
		if (b->data=malloc(b->nb=n)) {
			return b;
		}
		free(b);
		return 0;
	}
	return 0;
}
static
freeb(Msg *b)
{
	free(b->data);
	free(b);
}
static
sleepq(Queue *q, Proc *p)
{
	Proc *prev,*cur;
	p->next=0;
	for (prev=0,cur=q->waiting; cur;cur=cur->next);
	if (prev==0)
		q->waiting = p;
	else
		prev->next = p;
}


static Msg *
getq(Queue *q)
{
	Msg *b;
	if ((b=q->head) && (q->head=b->next)==0) {
		q->tail=0;
	}
	return b;
}

static
putq(Queue *q, Msg *b)
{
	if (q->waiting) {
		Proc *p=q->waiting;
		q->waiting=p->next;
		reply_read(p->pid, b->data, b->nb);
		free(p);
		freeb(b);
	} else {
		b->next=0;
		if (q->tail) {
			q->tail->next = b;
			q->tail = b;
		} else {
			q->head=q->tail=b;
		}
	}
}


static
pipe_write(pid_t p, file_t *f, unsigned n, void *buf, unsigned len)
{
	struct _io_write_reply w;
	Msg     *b;
	Queue   *q = f->inode->ino_data;

	if (len>n)
		len = n;
	if ((b=allocb(n))==0) {
		w.status = ENOMEM;
		Reply(p, &w.status, sizeof w.status);
		return;
	}
	memcpy(b->data, buf, len);
	if (n>len) {
		unsigned t;
		if ((t=wcopyin(p, len, b->data+len, n-len)) != n-len) {
			w.status = errno;
			Reply(p, &w.status, sizeof w.status);
			freeb(b);
			return;
		}
	}
	putq(q,b);
	w.status = EOK;
	w.nbytes = n;
	w.zero  = 0;
	Reply(p, &w, sizeof w);
}

static
pipe_read(pid_t p, file_t *f, long nbytes)
{
	Msg     *b;
	Queue   *q = f->inode->ino_data;
	if ((b=getq(q))==0) {
		Proc   *P;
		if (nonblock(f)) {
			reply_errno(p, EWOULDBLOCK);
			return;
		}
		if ((P=malloc(sizeof *P)) == 0) {
			reply_errno(p, ENOMEM);
			return;
		}
		P->pid = p;
		P->size = nbytes;
		sleepq(q, P);
	} else {
		reply_read(p, b->data, b->nb);
		freeb(b);
	}
	return;
}

proc_killed(int pid)
{
	Queue *q;
	for (q=Qlist; q; q=q->qnext) {
		Proc   *p,*c;
		for (p=0,c=q->waiting; c&&c->pid!=pid;p=c,c=c->next);
		if (c) {
			if (p)
				p->next=c->next;
			else
				q->waiting=c->next;
			free(c);
			reply_errno(pid, EINTR);
			return 0;
		}
	}
	return 1;
}


create_dev(inode_t **ip)
{
	inode_t     *io;
	Queue       *q;
	static   iofuns pclass;

	if (!pclass.write) {
		pclass.write = pipe_write;
		pclass.read  = pipe_read;
	}

	if ((q=malloc(sizeof *q)) == 0)
		return ENOMEM;

	q->head=q->tail = 0;
	q->waiting = 0;

	if ((io=inode_alloc(&pclass)) == 0) {
		free(q);
		return ENOMEM;
	}
	pclass.minor++;
	io->ino_ftype |= S_IFCHR;
	io->ino_data = q;
	q->qnext = Qlist;
	Qlist = q;
	*ip = io;
	return EOK;
}
