/*-
 * This one handles reading and writing but also demonstrates how a
 * client can request different types of information at different times
 * using only read() and write().
 *
 * With this example clients can write two different types of messages:
 *
 *   "SET VOLTAGE=100"
 *   "GET VOLTAGE"
 *
 * When the volt_write() function gets the "SET VOLTAGE=" message it stores
 * away the voltage that is after the equal sign ("=").  In the above
 * example that is 100.
 *
 * When the volt_write() function gets the "GET VOLTAGE" message it makes
 * a note that the writer had requested this.  The next time that
 * writer does a read(), the volt_read() function will return with the
 * voltage, "VOLTAGE=100".
 *
 */

#include <libc.h>
#include "iomgr.h"

#include <sys/io_msg.h>
#include <sys/kernel.h>
#include <sys/uio.h>

#define REQUESTING_VOLTAGE     1
// define other request types here

typedef struct Data  Data;


// One 'Data' per device, contains device level information such as the
// voltage.  requestq is a list of processes who will shortly be doing a
// read along with what they will be expecting from the read.
struct Data {
    float       voltage;
};

extern int  verbose;
extern char *progname;

static
volt_write(pid_t p, file_t *f, unsigned n, void *buf, unsigned len)
{
	struct _io_write_reply  w;
	char                    *b;
	Data    *d = f->inode->ino_data;

    if (verbose) printf("%s: Inside volt_write()\n", progname);
	if (len>n)
		len = n;
	if ((b=malloc(n))==0) {
		w.status = ENOMEM;
		Reply(p, &w.status, sizeof w.status);
		return;
	}
	memcpy(b, buf, len);
	if (n>len) {
		unsigned t;
		if ((t=wcopyin(p, len, b+len, n-len)) != n-len) {
			w.status = errno;
			Reply(p, &w.status, sizeof w.status);
			free(b);
			return;
		}
	}
    if (verbose) printf("%s: Received \"%s\"\n", progname, b);

    if (!strnicmp(b, "SET VOLTAGE=", 12)) 
    {
	/* Copy the voltage value to f->inode->ino_data */
	d->voltage = atof(strchr(b, '=')+1);
        if (verbose) printf("%s: Did %s\n", progname, b);
    } 
    else if (!stricmp(b, "GET VOLTAGE")) 
    {
	/* Set indicator for requesting data */
	f->f_data[0] = REQUESTING_VOLTAGE;
    }

    w.status = EOK;
    w.nbytes = n;
    w.zero  = 0;
    Reply(p, &w, sizeof w);
    free(b);
}

static
volt_read(pid_t p, file_t *f, long nbytes)
{
	char    b[30];
	Data    *d = f->inode->ino_data;

    if (verbose) printf("%s: Inside volt_read(%d,...)\n", progname, p);

    // check if this process is requesting anything
    
    
	
    switch (f->f_data[0]) 
    {
        case REQUESTING_VOLTAGE:
            sprintf(b, "VOLTAGE=%f", d->voltage);
            if (verbose) printf("%s: Replying with %s\n", progname, b);
    		reply_read(p, b, strlen(b)+1);
            break;
	default:
           // this process had not previously called write() in order to
           // request anything
           if (verbose) printf("%s: Replying ENOENT\n", progname);
	   reply_errno(p, ENOENT);
	break;
    }
}

proc_killed(int pid)
{
    
            if ( verbose )
                printf( "%s: Inside proc_killed(), replying with EINTR\n",
                        progname );
            reply_errno(pid, EINTR);
            return 0;
}

create_dev(inode_t **ip)
{
	inode_t         *io;
    Data            *d;
	static iofuns   vclass;

	if (!vclass.write) {
		vclass.write = volt_write;
		vclass.read  = volt_read;
	}

	if ((d=malloc(sizeof *d)) == 0)
		return ENOMEM;

	if ((io=inode_alloc(&vclass)) == 0) {
		return ENOMEM;
	}
	vclass.minor++;
	io->ino_ftype |= S_IFCHR;
	io->ino_data = d;
	*ip = io;
	return EOK;
}
