/*-
 * This one handles reading and writing but also demonstrates how a
 * client can request different types of information at different times
 * using only read() and write().
 *
 * With this example clients can write two different types of messages:
 *
 *   "SET VOLTAGE=100"
 *   "GET VOLTAGE"
 *
 * When the volt_write() function gets the "SET VOLTAGE=" message it stores
 * away the voltage that is after the equal sign ("=").  In the above
 * example that is 100.
 *
 * When the volt_write() function gets the "GET VOLTAGE" message it makes
 * a note that the writer had requested this.  The next time that
 * writer does a read(), the volt_read() function will return with the
 * voltage, "VOLTAGE=100".
 *
 */

#include <libc.h>
#include "iomgr.h"

#include <sys/io_msg.h>
#include <sys/kernel.h>
#include <sys/uio.h>

#define REQUESTING_VOLTAGE     1
// define other request types here

typedef struct Proc  Proc;
typedef struct Data  Data;

static struct Data  *dList;

// One 'Data' per device, contains device level information such as the
// voltage.  requestq is a list of processes who will shortly be doing a
// read along with what they will be expecting from the read.
struct Data {
    float       voltage;
    Proc        *requestq;
    Data        *next;
};

struct Proc {
	pid_t       pid;
	Proc        *next;
	int         requesting; // see REQUESTING_* defines above
};

extern int  verbose;
extern char *progname;

static
addToWaitingQueue(Data *d, Proc *p, pid_t pid, int type)
{
	Proc *prev, *cur;

	p->next=0;
	for (prev=0, cur=d->requestq; cur; cur=cur->next)
        ;
	if (prev==0)
		d->requestq = p;
	else
		prev->next = p;
    p->pid        = pid;
    p->requesting = type;
}

static
volt_write(pid_t p, file_t *f, unsigned n, void *buf, unsigned len)
{
	struct _io_write_reply  w;
	char                    *b;
    Data                    *d = f->inode->ino_data;
    Proc                    *requester;

    if (verbose) printf("%s: Inside volt_write()\n", progname);
	if (len>n)
		len = n;
	if ((b=malloc(n))==0) {
		w.status = ENOMEM;
		Reply(p, &w.status, sizeof w.status);
		return;
	}
	memcpy(b, buf, len);
	if (n>len) {
		unsigned t;
		if ((t=wcopyin(p, len, b+len, n-len)) != n-len) {
			w.status = errno;
			Reply(p, &w.status, sizeof w.status);
			free(b);
			return;
		}
	}
    if (verbose) printf("%s: Received \"%s\"\n", progname, b);
    if (!strnicmp(b, "SET VOLTAGE=", 12)) {
        d->voltage = atof(strchr(b, '=')+1);
        if (verbose) printf("%s: Did %s\n", progname, b);
    } else if (!stricmp(b, "GET VOLTAGE")) {
        if ((requester = malloc(sizeof(*requester))) == 0) {
			w.status = errno;
			Reply(p, &w.status, sizeof w.status);
			free(b);
			return;
        }
        if (verbose) printf("%s: Did %s\n", progname, b);
        addToWaitingQueue(d, requester, p, REQUESTING_VOLTAGE);
    }
	w.status = EOK;
	w.nbytes = n;
	w.zero  = 0;
	Reply(p, &w, sizeof w);
    free(b);
}

static
volt_read(pid_t p, file_t *f, long nbytes)
{
	char    b[30];
	Data    *d = f->inode->ino_data;
    Proc    *r, *c;

    if (verbose) printf("%s: Inside volt_read(%d,...)\n", progname, p);

    // check if this process is requesting anything
    
    for (r=0, c=d->requestq; c && c->pid != p; r=c, c=c->next);
    
    if (c) {
        switch (c->requesting) {
        case REQUESTING_VOLTAGE:
            sprintf(b, "VOLTAGE=%f", d->voltage);
            if (verbose) printf("%s: Replying with %s\n", progname, b);
    		reply_read(p, b, strlen(b)+1);
            if (r)
                r->next=c->next;
            else
                d->requestq=c->next;
            free(c);
            break;
        }
    } else {
        // this process had not previously called write() in order to
        // request anything
        if (verbose) printf("%s: Replying ENOENT\n", progname);
		reply_errno(p, ENOENT);
    }
}

proc_killed(int pid)
{
	Proc *r, *c;
	Data *d;
    
	for (d=dList; d; d=d->next) {
        
        // check if this process is requesting anything

        for (r=0,c=d->requestq; c&&c->pid!=pid;r=c,c=c->next);

        if (c) {
            if (r)
                r->next=c->next;
            else
                d->requestq=c->next;
            free(c);
            if ( verbose )
                printf( "%s: Inside proc_killed(), replying with EINTR\n",
                        progname );
            reply_errno(pid, EINTR);
            return 0;
        }
	}
	return 1;
}

create_dev(inode_t **ip)
{
	inode_t         *io;
    Data            *d;
	static iofuns   vclass;

	if (!vclass.write) {
		vclass.write = volt_write;
		vclass.read  = volt_read;
	}

	if ((d=malloc(sizeof *d)) == 0)
		return ENOMEM;

	d->requestq = 0;

	if ((io=inode_alloc(&vclass)) == 0) {
		return ENOMEM;
	}
	vclass.minor++;
	io->ino_ftype |= S_IFCHR;
	io->ino_data = d;
    d->next = dList;
    dList = d;
	*ip = io;
	return EOK;
}
