/*-
 * io: a framework for creating IO managers.
 */
#ifndef io_h
#define io_h

#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/fd.h>


#include "externs.h"


#define gettime()    __time(0)

#define _offof(x,y)    (((char *)&(x.y))-((char *)&(x)))
#define USED(x)    {(x)=(x);}


typedef struct inode  inode_t;
typedef struct file   file_t;
typedef struct iofuns iofuns;
typedef pid_t proc_t;

typedef int err_t;    /* returns errno */


enum {
	FNONE  = -1,
	FEXEC  = (1 << 0),
	FWRITE = (1 << 1),
	FREAD  = (1 << 2),
	FACC   = (FREAD|FWRITE|FEXEC),
	FOWNER = (1 << 3),
	FNDELAY= (1 << 4),
	FSYNC  = (1 << 5)
};

struct file {
	mode_t		mode;
	int		count;
	off_t		offset;
	inode_t        *inode;
	void           *fdata; /* interesting to driver */
};


struct ino {
    ino_t           ino_ino;         /*  File serial number. */
    dev_t           ino_dev;         /*  ID of device containing file. */
    mode_t          ino_ftype;       /*  type and permissions */
    off_t           ino_size;        /*  file size */
    dev_t           ino_rdev;        /*  Device ID, for inode that is device */
    short           ino_uid;
    short           ino_gid;
    time_t          ino_mtime,       /*  Time of last data modification      */
                    ino_atime,       /*  Time last accessed                  */
                    ino_ctime;       /*  Time of last status change          */
    nlink_t         ino_nlink;
};

extern void ino2stat(struct stat *st, struct ino *ino);


/*- inode.h: generic io device */
struct inode {
/* private: */
	struct ino    i_ino;
	void         *data;	/* interesting to driver */
/* protected: */
	long          devno;
/* public: */
	iofuns       *driver;
};
#define ino_init(i)   ((i)->driver->init)
#define ino_read(i)   ((i)->driver->read)
#define ino_write(i)  ((i)->driver->write)
#define ino_access(i) ((i)->driver->access)
#define ino_open(i)   ((i)->driver->close)
#define ino_ioctl(i)  ((i)->driver->ioctl)
#define ino_link(i)   ((i)->driver->link)
#define ino_unlink(i)   ((i)->driver->unlink)
#define ino_lookup(i) ((i)->driver->lookup)
#define ino_creat(i)  ((i)->driver->creat)
#define ino_setsize(i)  ((i)->driver->setsize)

struct iofuns {
	int         (*init)();
	int         (*lookup)();
	int         (*read)();
	int         (*write)();
	int         (*access)();
	int         (*open)();
	int         (*close)();
	int         (*ioctl)();
	int         (*link)();
	int         (*unlink)();
	int         (*creat)();
	int         (*setsize)();
};

/*-
 *   less cumbersome macroes:
 */
extern err_t d_init(inode_t *d, unsigned major);
extern err_t d_lookup(inode_t *dir, proc_t p, char *name, inode_t **res);
extern err_t d_link(inode_t *dir, char *name, inode_t *file);
extern err_t d_unlink(inode_t *dir, char *name);
extern err_t d_creat(inode_t *, int perm, uid_t u, gid_t g, inode_t **ip);
extern err_t d_setsize(inode_t *, ulong_t len);

extern err_t f_open(proc_t p, file_t *f, unsigned mode);
extern err_t f_close(proc_t p, file_t *f);
extern int   f_read(proc_t p, file_t *f, unsigned n);
extern int   f_write(proc_t p, file_t *f, unsigned n);

#define d_init(d,m)       (*ino_init(d))((d),(m))
#define d_lookup(d,p,n,r) (*ino_lookup(d))((d),(p),(n),(r))
#define d_link(d,n,f)     (*ino_link(d))((d),(n),(f))
#define d_unlink(d,n)     (*ino_unlink(d))((d),(n))
#define d_creat(d,p,u,g,ip) (*ino_creat(d))((d),(p),(u),(g),(ip))
#define d_setsize(d,len) (*ino_setsize(d))((d),(len))

#define f_open(p,f,m)     (*ino_open(f->inode))((p),(f),(m))
#define f_read(p, f, n)   (*ino_read(f->inode))((p),(f),(n))
#define f_write(p, f, n)  (*ino_write(f->inode))((p),(f),(n))
#define f_close(p,f)      (*ino_close(f->inode))((p),(f))

#define d_size(d)	  (d->i_ino.ino_size)
#define d_owner(d)	  (d->i_ino.ino_uid)
#define d_group(d)        (d->i_ino.ino_gid)
#define d_mode(d)         (d->i_ino.ino_ftype)


/*-
 * Macroes checking various open modes of a file,
 * readok(f) 	- non zero if readable
 * writeok(f) 	- non zero if writable
 * isowner(f) 	- non zero if effective owner perms
 * nonblock(f) 	- non zero if opened O_NONBLOCK.
 */

#define readok(f)   ((f)->mode & FREAD)
#define writeok(f)  ((f)->mode & FWRITE)
#define isowner(f)  ((f)->mode & FOWNER)
#define nonblock(f) ((f)->mode & FNDELAY)
#define issync(f)   ((f)->mode & FSYNC)



/*-
 * system interface: functions extract bits of OS information.
 */
extern file_t *lookup_fd(proc_t pid, int fd, int errok);
extern file_t *falloc(void);
extern void    ffree(file_t *);
extern int     f2index(file_t *f);
extern         ids(pid_t pid, uid_t *u, gid_t *g);
extern int     isroot(pid_t pid);
extern err_t   chk_perm(uid_t u, gid_t g, mode_t mode, inode_t *obj);


/*-
 * io-class messages.
 */
extern int io_open(proc_t pid, mode_t mode, char *path, int fd, int unit, int perm);
extern int io_chdir(proc_t pid, char *path, int unit);
extern int io_handle(proc_t pid, mode_t mode, char *path, int fd, int unit, int perm);
extern int io_stat(proc_t pid, char *path, int unit);
extern int io_fstat(proc_t pid, file_t *f);
extern int io_chmod(proc_t pid, file_t *f, mode_t mode);
extern int io_chown(proc_t pid, file_t *f, uid_t uid, gid_t gid);
extern int io_close(proc_t pid, file_t *f);
extern int io_read(proc_t pid, file_t *f, int nbytes);
extern int io_write(proc_t pid, file_t *f, int nbytes);
extern int io_dup(proc_t pid, file_t *f, proc_t dstpid, int dstfd);
extern int io_utime(proc_t pid, file_t *f, time_t mtime, time_t atime);
extern int io_lseek(proc_t pid, file_t *f, int whence, off_t loc);
extern int io_rewinddir(proc_t pid, file_t *f);
extern int io_readdir(proc_t pid, file_t *f, int ndirs);
/*-
 * fs-class messages
 */
extern int fs_mkdir(pid_t pid, char *path, int unit, mode_t modes);
extern int fs_unlink(pid_t pid, char *path, int unit);



extern int namei(proc_t pid, char *path, mode_t mode, inode_t *root, inode_t **res);
extern int nameip(proc_t pid, char *path, mode_t mode, inode_t *root,
                 inode_t **pptr, char *component);
extern int _open(pid_t pid, mode_t mode, inode_t *p, int fd);
extern mode_t base_modes(mode_t t);


/*-
 * some assist for finishing messages.
 */
extern int reply_errno(int pid, int stat);
/*- convenient interface */
#define GetMsg(p,m)  Readmsg(p, 0, &m, sizeof m)
#define PutMsg(p,m)  Reply(p, &m, sizeof m)
extern int reply_read(int pid, void *mem, int n);
extern int reply_readv(int pid, int n, struct iovec *v);
extern int FD_SET(int pid, int fd, int handle);

#if 0
#define FD_SET(p,f,x)  qnx_fd_attach((p), (f), 0, 0, 0, 0, (x))
#endif

#define FD_CLEAR(p,f)  FD_SET((p), (f), 0)

extern inode_t *inode_alloc(iofuns *class);
extern int      inode_init(int ndevs);

extern int      root_mount(const char *basedir);
extern inode_t *getroot(int unitno);


#endif
