/*-
 * dir: manage a special device type of directory.
 * I'm not sure whether it was a mistake to have directories
 * be 'normal' device type; for now waffle.
 * strangeness in size -- the directory size is reported as
 * the number of entries/sizeof dirent for consistency.
 *
 */

#include <errno.h>
#include <stddef.h>
#include <dirent.h>
#include <string.h>
#include <stdlib.h>
#include <sys/kernel.h>
#include <sys/io_msg.h>

#include "iointernal.h"

#define DIR_READ_SIZE sizeof(struct dirent)
static int maxdev = 64;

typedef struct directory directory;
struct directory {
	char     dir_name[NAME_MAX+1];
	inode_t *dir_ino;
};
/* bytes to dirs */
#define NDIR(o)  (o/sizeof(directory))

static
dir_lookup(inode_t *dir, char *name, inode_t **result)
{
	int        i;
	unsigned   nents = NDIR(dir->ino_size);
	directory *pdir = dir->ino_data;
	for (i=0; i < nents; i++) {
		if (pdir[i].dir_ino && strcmp(pdir[i].dir_name, name) == 0)
			break;
	}
	if (i == nents)
		return ENOENT;
	*result = pdir[i].dir_ino;
	return EOK;
}

static
dir_link(inode_t *dir, char *name, inode_t *ent)
{
	unsigned   nents = NDIR(dir->ino_size);
	directory *pdir  = dir->ino_data;
	unsigned   i;

	for (i=0; i < nents && pdir[i].dir_ino; i++);
	if (i == nents) {
		return ENOSPC;
	}
	ent->ino_dev = dir->ino_dev;
	pdir[i].dir_ino = ent;
	strcpy(pdir[i].dir_name,name);
	pdir[i].dir_ino->ino_nlink++;
	return EOK;
}

static
dir_unlink(inode_t *dir, char *name)
{
	unsigned   nents = NDIR(dir->ino_size);
	directory *pdir  = dir->ino_data;
	inode_t   *ent;
	unsigned   i;

	for (i=0, ent=0; i < nents ; i++) {
		if (pdir[i].dir_ino && strcmp(pdir[i].dir_name, name)==0) {
			ent = pdir[i].dir_ino;
			break;
		}
	}
	if (i == nents) {
		return EINVAL;
	}
	pdir[i].dir_ino = 0;
	ent->ino_nlink--; /* might close/exit it! */
	return EOK;
}


/*-
 * Note: This routine actually responds with a readdir reply.
 * inexplicably, read() and readdir() use different messages.
 */

static
dir_read(pid_t pid, file_t *f, long nbytes)
{
	struct _io_read_reply rep;
	int                   repoffs = _offof(rep, data);
	int                   nents = nbytes / DIR_READ_SIZE;
	int                   base  = NDIR(f->offset);
	inode_t              *inop = f->inode;
	directory            *dirp = inop->ino_data;

	rep.zero  = 0;
	while (f->offset < inop->ino_size && nents) {
		inode_t       *p;
	/*-
	 * Grossly inefficient.  Should have a bundle
	 * of em here and buffer them up.
	 */
	 	if (p=dirp[base].dir_ino) {
			struct dirent  dir;
	 		ino2stat(&dir.d_stat, p);
			strcpy(dir.d_name, dirp[base].dir_name);
			dir.d_stat.st_status |= _FILE_USED;
			Writemsg(pid, repoffs, &dir, sizeof dir);
			repoffs += sizeof dir;
			nents--;
		}
		base++;
		f->offset += sizeof dirp[0];
	}
	rep.nbytes = (nbytes/DIR_READ_SIZE) - nents;
	rep.status = EOK;
	return Reply(pid, &rep, offsetof(struct _io_read_reply, data));
}

static
dir_write(pid_t p, file_t *f, long nbytes)
{
	USED(f);
	USED(nbytes);
	return reply_errno(p, EPERM);
}

/*-
 * what to do with them...
 */
static
dir_open(proc_t p, file_t *f, unsigned mode)
{
	return iodef_open(p, f, mode);
}

static
dir_close(file_t *f)
{
	return iodef_close(f);
}


inode_t *
create_dir(int creat(inode_t **))
{
	inode_t     *io;
	directory   *dir;
	static iofuns dirclass;

	if (!dirclass.open) {
		/*- until C has tagged initializers */
		dirclass.open  = dir_open;
		dirclass.close = dir_close;
		dirclass.lookup = dir_lookup;
		dirclass.read  = dir_read;
		dirclass.write = dir_write;
		dirclass.link  = dir_link;
		dirclass.unlink= dir_unlink;
		dirclass.creat = creat;
	}

	if ((dir=calloc(sizeof(*dir), maxdev)) == 0)
		return 0;

	if ((io=inode_alloc(&dirclass)) == 0) {
		free(dir);
		return 0;
	}
	io->ino_size = maxdev * sizeof *dir;
	io->ino_ftype = S_IFDIR | 0775;
	io->ino_data            = dir;
	return io;
}
