/*-
 * io: a framework for creating IO managers.
 */
#ifndef iomgr_h
#define iomgr_h

#include <sys/types.h>
#include <sys/uio.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>

#ifdef IN_PROC

#include "externs.h"
#define gettime()    __time(0)

#else
#include <time.h>
#define gettime() time(0)

#endif


#define _offof(x,y)    (((char *)&(x.y))-((char *)&(x)))
#define USED(x)    {(x)=(x);}


typedef struct inode  inode_t;
typedef struct file   file_t;
typedef struct iofuns iofuns;
typedef pid_t proc_t;

typedef int err_t;    /* returns errno */


enum {
	FNONE  = 0,
	FEXEC  = (1 << 0),
	FWRITE = (1 << 1),
	FREAD  = (1 << 2),
	FACC   = (FREAD|FWRITE|FEXEC),
	FOWNER = (1 << 3),
	FNDELAY= (1 << 4),
	FSYNC  = (1 << 5)
};

struct file {
	unsigned        mode;
	unsigned	count;
	long 		offset;
	inode_t        *inode;
	ulong_t         f_data[4]; /* scratch area for driver */
};


extern void ino2stat(struct stat *st, inode_t *ino);

struct iofuns {
	int         major;
	int         minor;
	int         (*open)();
	int         (*close)();
	int         (*read)();
	int         (*write)();
	int         (*link)();      /* attach a new member */
	int         (*unlink)();    /* remove a member */
	int         (*lookup)();    /* scan for a member */
	int         (*creat)();     /* build a new one */
	int         (*setsize)();   /* change the size O_TRUNC, ltrunc()*/
	int         (*ioctl)();     /* not used */
	int         (*qioctl)();     /* not used */
};

/*- inode.h: generic io device */
struct inode {
	ino_t           ino_ino;         /*  File serial number. */
	dev_t           ino_dev;         /*  ID of device containing file. */
	mode_t          ino_ftype;       /*  type and permissions */
	off_t           ino_size;        /*  file size */
	dev_t           ino_rdev;        /*  Device ID, for inode that is device */
	short           ino_uid;
	short           ino_gid;
	time_t          ino_mtime,       /*  Time of last data modification      */
			ino_atime,       /*  Time last accessed                  */
			ino_ctime;       /*  Time of last status change          */
	nlink_t         ino_nlink;
	void           *ino_data;	/* interesting to driver */
	iofuns         *ino_driver;
};


#define d_size(d)	  (d->ino_size)
#define d_owner(d)	  (d->ino_uid)
#define d_group(d)        (d->ino_gid)
#define d_mode(d)         (d->ino_ftype)


/*-
 * Macroes checking various open modes of a file,
 * readok(f) 	- non zero if readable
 * writeok(f) 	- non zero if writable
 * isowner(f) 	- non zero if effective owner perms
 * nonblock(f) 	- non zero if opened O_NONBLOCK.
 */

#define readok(f)   ((f)->mode & FREAD)
#define writeok(f)  ((f)->mode & FWRITE)
#define isowner(f)  ((f)->mode & FOWNER)
#define nonblock(f) ((f)->mode & FNDELAY)
#define issync(f)   ((f)->mode & FSYNC)


/*-
 * system interface: functions extract bits of OS information.
 */
extern file_t *lookup_fd(proc_t pid, int fd, int errok);
extern file_t *falloc(void);
extern void    ffree(file_t *);
extern int     f2index(file_t *f);
extern file_t *index2f(int index);
extern         ids(pid_t pid, uid_t *u, gid_t *g);
extern int     isroot(pid_t pid);
extern err_t   chk_perm(uid_t u, gid_t g, mode_t mode, inode_t *obj);


/*-
 * some assist for finishing messages.
 */
#define GetMsg(p,m)  Readmsg(p, 0, &m, sizeof m)
#define PutMsg(p,m)  Reply(p, &m, sizeof m)
extern int      reply_errno(int pid, int stat);
extern int      reply_read(int pid, void *mem, int n);
extern int      reply_readv(int pid, int n, struct iovec *v);
extern int      FD_SET(int pid, int fd, int handle);

extern int      wcopyin(proc_t p, ulong_t off, char *buf, ulong_t n);
extern int      rcopyout(proc_t p, ulong_t off, char *buf, ulong_t n);

extern inode_t *inode_alloc(iofuns *class);
extern int      inode_init(int ndevs);

extern int      getunit(int index);
extern int      root_mount(const char *basedir, inode_t *root);
extern inode_t *getroot(int unitno);

extern inode_t *create_dir(int creat(inode_t **ip));

extern          server_addpid(pid_t p, int (*f)());
extern void     server(void);
extern int      server_init(void);

#endif
