
/*- server: main loop and message dispatch */

#include <dirent.h>
#include <string.h>
#include <sys/kernel.h>
#include "iointernal.h"
#include "iomessage.h"

#define UBITS 07777


err_t
_fname(pid_t pid, struct _io_open *o, inode_t **r, char *path, int len)
{
	inode_t *root;
	int      n;
	if ((root=getroot(o->unit)) == 0) {
		return ENOENT;
	}
	if (len < _offof((*o),path[PATH_MAX])) {
		if ((n=Readmsg(pid, _offof((*o), path), path, PATH_MAX)) == -1) {
			return EFAULT;
		}
		path[n] = '\0';
	} else {
		strncpy(path, o->path, PATH_MAX);
	}
	*r = root;
	return EOK;
}

io_message(pid_t pid, io_msg *m, unsigned len)
{
	file_t        *f;
	inode_t       *root=0;
	int            e;
	char           path[PATH_MAX+1];

	/*- all have a variable path component; read it in to simplify
	 *  namei()
	 */
	switch (m->type) {
	case _IO_HANDLE:
		if ((e=_fname(pid, &m->open, &root, path, len)) != EOK) {
			reply_errno(pid, e);
			break;
		}
		io_handle(pid, m->open.oflag, path, m->open.fd, root, m->open.eflag);
		break;

	case _IO_OPEN:
		if ((e=_fname(pid, &m->open, &root, path, len)) != EOK) {
			reply_errno(pid, e);
			break;
		}
		io_open(pid, m->open.oflag, path, m->open.fd, root, m->open.mode);
		break;

	case _IO_FLAGS:
		if ((e=_fname(pid, &m->open, &root, path, len)) != EOK) {
			reply_errno(pid, e);
			break;
		}
		io_flags(pid, path, root);
		break;
#ifdef NO_FLAGS
		reply_errno(pid, ENOSYS);
		break;
#endif

	case _IO_LOCK:
		reply_errno(pid, ENOSYS);
		break;

	case _IO_CHDIR:
		if ((e=_fname(pid, &m->open, &root, path, len)) != EOK) {
			reply_errno(pid, e);
			break;
		}
		io_chdir(pid, path, root);
		break;

	case _IO_STAT:
		if ((e=_fname(pid, &m->open, &root, path, len)) != EOK) {
			reply_errno(pid, e);
			break;
		}
		io_stat(pid, path, root);
		break;

	case _IO_RENAME:
		reply_errno(pid, ENOSYS);
		break;

	case _IO_GET_CONFIG:
		reply_errno(pid, ENOSYS);
		break;

	case _IO_DUP:
		/*- should check for forgery or loader thread */
		if (m->dup.src_pid == 0) {
			m->dup.src_pid = pid;
		}
		if ((f=lookup_fd(m->dup.src_pid, m->dup.src_fd, 0)) == 0)
			break;
		io_dup(pid, f, m->dup.src_pid, m->dup.dst_fd);
		break;

	case _IO_CLOSE:
		if ((f=lookup_fd(pid, m->close.fd, 0)) == 0) {
			break;
		}
		if(--f->count == 0) {
			f_close(pid, f);
			ffree(f);
		}
		reply_errno(pid, EOK);
		break;

	case _IO_FSTAT:
		if ((f=lookup_fd(pid, m->fstat.fd, 0)) == 0)
			break;
		io_fstat(pid, f);
		break;

	case _IO_CHMOD:
		if ((f=lookup_fd(pid, m->chmod.fd, 0)) == 0)
			break;
		if (isowner(f)==0) {
			reply_errno(pid, EBADF);
		}
		f->inode->ino_ftype &= ~UBITS;
		f->inode->ino_ftype |= m->chmod.mode&UBITS;
		reply_errno(pid, EOK);
		break;
			
	case _IO_CHOWN:
		if ((f=lookup_fd(pid, m->chown.fd, 0)) == 0)
			break;
		io_chown(pid, f, m->chown.uid, m->chown.gid);
		break;

	case _IO_UTIME:
		if ((f=lookup_fd(pid, m->utime.fd, 0)) == 0)
			break;
		if (!isowner(f)) {
			reply_errno(pid, EBADF);
			break;
		}
		f->inode->ino_mtime = m->utime.cur_flag ?
		                            gettime():m->utime.modtime;
		f->inode->ino_atime = m->utime.cur_flag ?
		                            gettime():m->utime.actime;
		reply_errno(pid, EOK);
		break;

	case _IO_LSEEK:
		if ((f=lookup_fd(pid, m->lseek.fd, 0)) == 0)
			break;
		io_lseek(pid, f, m->lseek.whence, m->lseek.offset);
		break;

	case _IO_REWINDDIR:
		if ((f=lookup_fd(pid, m->rewinddir.fd, 0)) == 0) {
			break;
		}
		if (!S_ISDIR(f->inode->ino_ftype)) {
			reply_errno(pid, EBADF);
			break;
		}
		io_lseek(pid, f, SEEK_SET, 0);
		break;

	case _IO_READ:
		if ((f=lookup_fd(pid, m->read.fd, 0)) == 0)
			break;
		if (!readok(f)) {
			reply_errno(pid, EBADF);
			break;
		}
		f_read(pid, f, m->read.nbytes);
		break;

	case _IO_READDIR:
		if ((f=lookup_fd(pid, m->readdir.fd, 0)) == 0)
			break;
		if (!readok(f) || !S_ISDIR(f->inode->ino_ftype)) {
			reply_errno(pid, EBADF);
			break;
		}
		f_read(pid, f, m->readdir.ndirs * sizeof(struct dirent));
		break;

	case _IO_WRITE:
		if ((f=lookup_fd(pid, m->write.fd, 0)) == 0)
			break;
		if (!writeok(f)) {
			reply_errno(pid, EBADF);
			break;
		}
		f_write(pid, f, m->write.nbytes, m->write.data,
			len - _offof(m->write, data));
		break;

	case _IO_SELECT:
		reply_errno(pid, ENOSYS);
		break;

	case _IO_QIOCTL:
		if ((f=lookup_fd(pid, m->qioctl.fd, 0)) == 0)
			break;
		f_qioctl(pid, f, m->qioctl.request, m->qioctl.nbytes, 
			m->qioctl.data, len - _offof(m->qioctl, data));
		break;

	default:
		reply_errno(pid, ENOSYS);
	}
	return 0;
}
